/******************************************************************************
 * File: PropertyObject.cpp
 * Date: 20.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "PropertyObject.h"
#include "PropertyExceptions.h"

void PropertyObject::serialize(SerializationContext& ctx) const
{
	// Serialize all properties
	ctx.beginGroup(PropertyVector::Serialization_Group, true);
	_properties.serialize(PropertyVector::Serialization_Element, ctx, [](const std::shared_ptr<PropertyBase> property) {
		return !property->hasFlag(PropertyBase::Flag::ReadOnly);
	});
	ctx.endGroup();
}

void PropertyObject::deserialize(DeserializationContext& ctx)
{
	// Deserialize all properties
	if (ctx.beginGroup(PropertyVector::Serialization_Group))
	{
		_properties.deserialize(PropertyVector::Serialization_Element, ctx, [this](const SettingsContainer& settings) -> std::shared_ptr<PropertyBase> {
			PropertyID id = settings[PropertyBase::Serialization_Value_ID].toString();
			auto property = _properties.selectByID(id);

			if (property && !property->hasFlag(PropertyBase::Flag::ReadOnly))
				return property;
			else
				return nullptr;
		});

		ctx.endGroup();
	}
}

void PropertyObject::removeProperty(PropertyID id)
{
	if (id != PropertyID::None)
	{
		auto prop = _properties.selectByID(id);

		if (prop)
			removeProperty(prop.get());
		else
			throw PropertyItemException{this, _EXCPT(QString{"Tried to remove a non-existing property (ID: %1)"}.arg(id))};
	}
}

void PropertyObject::removeProperty(const PropertyBase* prop)
{
	if (prop)
	{
		auto it = _properties.find(prop);

		if (it != _properties.cend())
			_properties.erase(it);
		else
			throw PropertyItemException{this, _EXCPT("Tried to remove a property not belonging to this item")};
	}
}

std::shared_ptr<PropertyBase> PropertyObject::_createProperty(PropertyID id, std::function<std::shared_ptr<PropertyBase>()> creator)
{
	if (id == PropertyID::None)
		throw std::invalid_argument{_EXCPT("id may not be empty")};

	if (_properties.selectByID(id))
		throw PropertyItemException{this, _EXCPT(QString{"A property with ID '%1' already exists"}.arg(id))};

	auto prop = creator();

	try {	// Propagate initialization errors to the caller
		prop->initProperty();
	}
	catch (...) {
		throw;
	}

	_properties.push_back(prop);
	return prop;
}
