/******************************************************************************
 * File: JsonSettingsCodec.cpp
 * Date: 27.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "JsonSettingsCodec.h"
#include "SerializationExceptions.h"
#include "util/StringUtils.h"

#include <QJsonValue>

void JsonSettingsCodec::encodeSettings(const SettingsContainer& settings, QTextStream& stream)
{
	encode(settings, stream, 0, true);
}

void JsonSettingsCodec::decodeSettings(SettingsContainer& settings, QTextStream& stream)
{
	settings.clear();

	try {
		decode(settings, stream);
	} catch (SerializationException& e) {
		throw SerializationException{_EXCPT(QString{"Invalid JSON document: %1"}.arg(e.what()))};
	}
}

void JsonSettingsCodec::encode(const SettingsContainer& settings, QTextStream& stream, int level, bool lastEntry) const
{
	// Begin a new Json object
	QString objectName = (!settings.getName().isEmpty() && level > 0 ? QString{"\"%1\": "}.arg(settings.getName()) : "");
	writeJsonLine(QString{"%1%2"}.arg(objectName).arg(settings.isArray() ? "[" : "{"), stream, level);

	// Write all children of the current container
	auto children = settings.children();

	for (unsigned int i = 0; i < children.size(); ++i)
	{
		const auto& child = children[i];
		bool lastChild = (i + 1 >= children.size() && settings.values().isEmpty());

		if (settings.isArray())
		{
			// Each array element is added as an individual object
			writeJsonLine(child->isArray() ? "[" : "{", stream, level + 1);
			encode(*child, stream, level + 2, true);
			writeJsonLine(QString{"%1%2"}.arg(child->isArray() ? "]" : "}").arg(!lastChild ? "," : ""), stream, level + 1);
		}
		else
			encode(*child, stream, level + 1, lastChild);
	}

	// Write all settings of the current container
	writeJsonValues(settings, stream, level + 1);

	// Close the Json object
	writeJsonLine(QString{"%1%2"}.arg(settings.isArray() ? "]" : "}").arg(!lastEntry ? "," : ""), stream, level);
}

void JsonSettingsCodec::decode(SettingsContainer& settings, QTextStream& stream) const
{
	// We first need to tokenize the given input stream
	auto tokens = readJsonDocument(stream);

	// Now parse the tokens using a recursive descent parser
	if (!tokens.empty())
		parseJsonDocument(settings, tokens);
}

void JsonSettingsCodec::writeJsonValues(const SettingsContainer& settings, QTextStream& stream, int level) const
{
	// Write out each value
	auto keys = settings.values().keys();

	for (int i = 0; i < keys.size(); ++i)
	{
		auto valueName = keys[i];
		writeJsonValue(valueName, settings.values()[valueName], stream, level, i + 1 >= keys.size());
	}
}

void JsonSettingsCodec::writeJsonValue(QString valueName, const QVariant& value, QTextStream& stream, int level, bool lastEntry) const
{
	QJsonValue jsonValue = QJsonValue::fromVariant(value);	// Use Qt's Json value class to properly handle the variant type
	QString formattedValue;

	switch (jsonValue.type())
	{
	case QJsonValue::Bool:
		formattedValue = jsonValue.toBool() ? "true" : "false";
		break;

	case QJsonValue::Double:
		formattedValue = QString{"%1"}.arg(jsonValue.toDouble());
		break;

	default:
		formattedValue = "\"" + StringUtils::encodeEscapeCharacters(jsonValue.toString()) + "\"";
		break;
	}

	writeJsonLine(QString{"\"%1\": %2%3"}.arg(valueName).arg(formattedValue).arg(!lastEntry ? "," : ""), stream, level);
}

void JsonSettingsCodec::writeJsonLine(const QString& text, QTextStream& stream, int level) const
{
	QString line = text;

	// Tabify the line
	if (level > 0)
		line.insert(0, QString{"\t"}.repeated(level));

	stream << line << "\n";
}

std::vector<JsonSettingsCodec::Token> JsonSettingsCodec::readJsonDocument(QTextStream& stream) const
{
	// Set up the various regular expressions for parsing the Json document
	struct TokenizerEntry
	{
		TokenizerEntry(QString regExp, TokenType type, bool greedy = true) : re{"^" + regExp}, tokenType{type} { re.setMinimal(!greedy); }

		QRegExp re;
		TokenType tokenType{TokenType::Undefined};
	};

	static const std::vector<TokenizerEntry> tokenizers{
		TokenizerEntry{R"("([^"\\]|\\["\\])*")", TokenType::String, false},
		TokenizerEntry{R"(\{)", TokenType::ObjectBegin},
		TokenizerEntry{R"(\})", TokenType::ObjectEnd},
		TokenizerEntry{R"(\[)", TokenType::ArrayBegin},
		TokenizerEntry{R"(\])", TokenType::ArrayEnd},
		TokenizerEntry{R"(:)", TokenType::Colon},
		TokenizerEntry{R"(,)", TokenType::Comma},
		TokenizerEntry{R"([^,]+)", TokenType::Value},
	};

	// Tokenize the entire document, line by line
	std::vector<Token> tokens;

	while (!stream.atEnd())
	{
		QString line = stream.readLine();

		while (!line.isEmpty())
		{
			// Remove leading whitespaces before parsing the line (so that new tokens always begin at position 0)
			if ((line = line.trimmed()).isEmpty())
				break;

			bool matchFound = false;

			// Try each tokenizer for a match; if none was found, the Json document has an invalid syntax
			for (const auto& tokenizer : tokenizers)
			{
				auto indexFound = tokenizer.re.indexIn(line);

				if (indexFound == 0)	// Matches must begin at the start of the line
				{
					int matchedLength = tokenizer.re.matchedLength();
					QString matchedString = line.left(matchedLength);

					if (tokenizer.tokenType == TokenType::String)
						matchedString = StringUtils::decodeEscapeCharacters(matchedString);

					tokens.emplace_back(Token{tokenizer.tokenType, matchedString});
					line.remove(0, matchedLength);

					matchFound = true;
					break;
				}
			}

			if (!matchFound)
				throw SerializationException{QString{"Invalid token '%1'"}.arg(line)};
		}
	}

	return tokens;
}

void JsonSettingsCodec::parseJsonDocument(SettingsContainer& settings, const std::vector<Token>& tokens) const
{
	// Parse the entire document consisting of a large Json object; note that the used grammar is somewhat limited and tailored for how settings containers are stored

	// Doc ::= Object
	ParseContext parseContext;
	parseJsonObject(settings, tokens, parseContext);
}

void JsonSettingsCodec::parseJsonObject(SettingsContainer& settings, const std::vector<Token>& tokens, ParseContext& parseContext) const
{
	// Object ::= '{' ObjectItems '}' | '{' '}'
	rdpExpect(TokenType::ObjectBegin, tokens, parseContext, [&settings, &parseContext](QString) { parseContext.pushChildSettings(settings); });	// Push a new container for the object on top

	if (!rdpAccept(TokenType::ObjectEnd, tokens, parseContext, nullptr, true))	// Check for an empty object
		parseJsonObjectItems(settings, tokens, parseContext);

	rdpExpect(TokenType::ObjectEnd, tokens, parseContext, [&parseContext](QString) { parseContext.popChildSettings(); });	// And pop it was we're finished with it
}

void JsonSettingsCodec::parseJsonObjectItems(SettingsContainer& settings, const std::vector<Token>& tokens, ParseContext& parseContext) const
{
	// ObjectItems ::= ObjectItem {',' ObjectItem}
	do {
		parseJsonObjectItem(settings, tokens, parseContext);
	} while (rdpAccept(TokenType::Comma, tokens, parseContext));
}

void JsonSettingsCodec::parseJsonObjectItem(SettingsContainer& settings, const std::vector<Token>& tokens, ParseContext& parseContext) const
{
	// ObjectItem ::= String ':' (String|Value|Array|Object)
	rdpExpect(TokenType::String, tokens, parseContext, [&parseContext](QString value) { removeEnclosingQuotes(value); parseContext.identifier = value; });	// Remember the last identifier
	rdpExpect(TokenType::Colon, tokens, parseContext);

	// Lambdas to propery set variant values from the Json value
	auto stringAttribtor = [&parseContext](QString value) { removeEnclosingQuotes(value); parseContext.setSettingsValue(value); };
	auto valueAttributor = [&parseContext](QString value) {
		if (value.compare("true", Qt::CaseInsensitive) == 0)
			parseContext.setSettingsValue(true);
		else if (value.compare("false", Qt::CaseInsensitive) == 0)
			parseContext.setSettingsValue(false);
		else if (value.indexOf('.') != -1)
			parseContext.setSettingsValue(value.toDouble());
		else
			parseContext.setSettingsValue(value.toInt());
	};

	if (rdpAccept(TokenType::String, tokens, parseContext, stringAttribtor))
	{
		// Nothing else to do
	}
	else if (rdpAccept(TokenType::Value, tokens, parseContext, valueAttributor))
	{
		// Nothing else to do
	}
	else if (rdpAccept(TokenType::ArrayBegin, tokens, parseContext, nullptr, true))
		parseJsonArray(settings, tokens, parseContext);
	else if (rdpAccept(TokenType::ObjectBegin, tokens, parseContext, nullptr, true))
		parseJsonObject(settings, tokens, parseContext);
}

void JsonSettingsCodec::parseJsonArray(SettingsContainer& settings, const std::vector<Token>& tokens, ParseContext& parseContext) const
{
	// Array ::= '[' Object {',' Object} ']' | '[' ']'
	rdpExpect(TokenType::ArrayBegin, tokens, parseContext, [&settings, &parseContext](QString) { parseContext.pushChildSettings(settings, true); });

	if (!rdpAccept(TokenType::ArrayEnd, tokens, parseContext, nullptr, true))	// Check for an empty array
	{
		do {
			// Handle the array elements which are encapsulated as individual objects
			SettingsContainer arrayElemContainer;
			parseContext.settingsStack.push(&arrayElemContainer);
			parseContext.identifier = "$_array_elem_$";
			parseJsonObject(settings, tokens, parseContext);
			parseContext.settingsStack.pop();

			// Extract the elements from their encapsulating objects
			for (auto arrayElem : arrayElemContainer.children("$_array_elem_$"))
			{
				if (arrayElem->children().size() == 1)	// There must be exactly one child object in the encapsulating object
					*parseContext.settingsStack.top() << std::move(*arrayElem->children().at(0));
				else
					throw SerializationException{"Invalid container object found"};
			}
		} while (rdpAccept(TokenType::Comma, tokens, parseContext));
	}

	rdpExpect(TokenType::ArrayEnd, tokens, parseContext, [&parseContext](QString) { parseContext.popChildSettings(); });
}

bool JsonSettingsCodec::rdpAccept(JsonSettingsCodec::TokenType token, const std::vector<Token>& tokens, ParseContext& parseContext, std::function<void(QString)> attributor, bool peek) const
{
	if (parseContext.currentToken < tokens.size() && tokens[parseContext.currentToken].type == token)
	{
		if (!peek)
		{
			// Call the attributor with the current token value
			if (attributor)
				attributor(tokens[parseContext.currentToken].value);

			parseContext.currentToken += 1;
		}

		return true;
	}
	else
		return false;
}

void JsonSettingsCodec::rdpExpect(JsonSettingsCodec::TokenType token, const std::vector<Token>& tokens, ParseContext& parseContext, std::function<void(QString)> attributor) const
{
	if (!rdpAccept(token, tokens, parseContext, attributor))
		throw SerializationException{QString{"Unexpected symbol type (wanted %1, got %2)"}.arg(static_cast<int>(token)).arg(static_cast<int>(!tokens.empty() ? tokens.back().type : TokenType::Undefined))};
}

void JsonSettingsCodec::removeEnclosingQuotes(QString& text)
{
	if (text.left(1) == "\"")
		text.remove(0, 1);

	if (text.right(1) == "\"")
		text.chop(1);
}

void JsonSettingsCodec::ParseContext::pushChildSettings(SettingsContainer& settings, bool isArray)
{
	// Create a new child container (or push the root container if the stack is empty)
	if (!settingsStack.empty())
	{
		auto childContainer = settingsStack.top()->createChild(identifier, isArray);
		settingsStack.push(childContainer);
	}
	else
		settingsStack.push(&settings);
}

void JsonSettingsCodec::ParseContext::setSettingsValue(const QVariant& value)
{
	if (!settingsStack.empty())
	{
		auto& currentSettings = *settingsStack.top();
		currentSettings[identifier] = value;
	}
}
